##############################################################################
# $Id: can.pl,v 1.2 1999/08/26 04:06:57 jheiss Exp $
##############################################################################
# Subroutines for checking if you can ping and rsh to a machine.  Useful for
# checking connectivity and permissions before trying to execute commands on
# a machine.
##############################################################################
# Copyright (C) 1999  Jason Heiss (jheiss@ofb.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##############################################################################
# Pre-RCS version control:
# created ??/??/97
# last modified 04/09/98
##############################################################################
# $Log: can.pl,v $
# Revision 1.2  1999/08/26 04:06:57  jheiss
# Added copyright and GNU license message.
#
# Revision 1.1  1999/08/09 02:57:05  jheiss
# Initial revision
#
##############################################################################

require 'facts.pl';

# Constants
$RSH_TIMEOUT = 60;

sub can_ping
{
	my $host = $_[0];

	unless ($host)
	{
		die "Invalid argument to \&can_ping";
	}

	my $os = get_os('localhost');
	my $osmajor = get_os_major_version('localhost');

	my $ping;
	if ($os eq 'SunOS' && $osmajor == 4)
	{
		$ping = '/usr/etc/ping';
	}
	elsif ($os eq 'SunOS' && $osmajor == 5)
	{
		$ping = '/usr/sbin/ping';
	}
	else
	{
		die "Unknown OS on localhost";
	}

	if (system("$ping $host >> /dev/null 2>&1"))
	{
		return $FALSE;
	}
	else
	{
		return $TRUE;
	}
}

sub can_rsh
{
	my $host = $_[0];

	unless ($host)
	{
		die "Invalid argument to \&can_rsh";
	}

	my $os = get_os('localhost');
	my $osmajor = get_os_major_version('localhost');

	my $ping;
	if ($os eq 'SunOS' && $osmajor == 4)
	{
		$rsh = '/usr/ucb/rsh';
	}
	elsif ($os eq 'SunOS' && $osmajor == 5)
	{
		$rsh = '/bin/rsh';
	}
	else
	{
		die "Unknown OS on localhost";
	}

	# Occasionally an rsh to a machine will just hang indefinately.
	# As such we set an alarm such that we can timeout the rsh after
	# some period of time.
	my $errcode = 0;
	eval
	{
		# A die inside of an eval block just exits the eval.
		local $SIG{ALRM} = sub { die "RSH TIMED OUT" };
		alarm $RSH_TIMEOUT;  # Set an alarm to timeout the rsh
		$errcode = system("$rsh $host echo >> /dev/null 2>&1");
		alarm 0;  # Cancel alarm
	};

	if ($@ && $@ =~ /RSH TIMED OUT/)
	{
		#print "rsh timed out\n";
		$errcode = 1;
	}

	if ($errcode)
	{
		return $FALSE;
	}
	else
	{
		return $TRUE;
	}
}

return 1;

