##############################################################################
# $Id: status.pl,v 1.3 1999/08/26 04:06:57 jheiss Exp $
##############################################################################
# Subroutines and constants that form our status file system.  The various
# scripts of the jumpstart process communicate the current status of all of
# the machines they are working on via a central status file.
##############################################################################
# Copyright (C) 1999  Jason Heiss (jheiss@ofb.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##############################################################################
# Pre-RCS version control:
# created 7/28/97
# last modified 03/09/98
##############################################################################
# $Log: status.pl,v $
# Revision 1.3  1999/08/26 04:06:57  jheiss
# Added copyright and GNU license message.
#
# Revision 1.2  1999/08/09 04:00:38  jheiss
# Fixed typo
#
# Revision 1.1  1999/08/09 03:15:10  jheiss
# Initial revision
#
##############################################################################

#############
# Libraries #
#############
require 'basedir.pl';
require 'inarray.pl';

##################
# Clean up $PATH #
##################

# This library of functions is used by the CGI script which generates
# the status web pages.  For security reasons we have that CGI script
# run with taint mode on which checks for insecure things.  One thing
# it doesn't like is a $PATH that isn't set to a known value.  So we
# set $PATH to the most basic here.  The only thing this library runs
# in a shell is cp so this isn't a big deal.
$ENV{PATH} = '/bin:/usr/bin';

#############
# Constants #
#############
$DEBUGLEVEL = 0;

$SUPERLOCKFILE = "$LOGBASEDIR/.lock/super.lock";
$LOCKFILE = "$LOGBASEDIR/.lock/status.lock";
$STATFILE = "$LOGBASEDIR/status";
$LOCKWAIT = 1;

# Various status codes
$NOTTOUCHED = 100;
$INSUFFINFO = 101;
$CONFIGNOPING = 102;
$CONFIGNORSH = 103;
$UNSUPPARCH = 104;
$UNSUPPOSVER = 105;
$IPWRONGHOST = 106;
$HOSTWRONGIP = 107;
$HOSTSNOENTRY = 108;
$ISNISSERVER = 109;
$BADNISSERV = 110;
$CONFIGRULESFAIL = 111;
$CONFIGDIFFRP = 112;
$CONFIGDNJ = 113;
$NOBOOTSERV = 115;
$NOINSTALLSERV = 116;
$ROGUERARP = 117;
$ROGUEBOOTPARAMS = 118;
$CONFIGSUCCESS = 200;
$STARTNOPING = 201;
$STARTNORSH = 202;
$USERLO = 204;
$STARTDIFFRP = 205;
$STARTDNJ = 206;
$ODBSERVER = 207;
$NEEDSBNI = 251;
$STARTED = 300;
$JSNOPING = 301;
$JSNORSH = 351;
$SUNINSTALL = 400;
$PFINSTALL = 500;
$PACKAGEINSTALL = 600;
$FINISHSCRIPT = 700;
$FINALBOOT = 800;
$JUMPFIN = 900;
$TJFAIL = 901;
$TNFAIL = 902;
$READY = 1000;

# Associate user-friendly messages with the status codes
%codestring = (
$NOTTOUCHED, 'Not touched',
$INSUFFINFO, 'Insufficient info to configure new or down machine',
$CONFIGNOPING, 'config was unable to ping this machine',
$CONFIGNORSH, 'config was unable to rsh to this machine',
$UNSUPPARCH, 'Machine has an unsupported architecture',
$UNSUPPOSVER, 'Machine has an unsupported OS version',
$IPWRONGHOST, 'IP address mapped to wrong hostname in NIS hosts map',
$HOSTWRONGIP, 'Hostname mapped to wrong IP address in NIS hosts map',
$HOSTSNOENTRY, 'No entry for the hostname or IP address of this machine in the NIS hosts map',
$ISNISSERVER, 'Machine is an approved NIS server',
$BADNISSERV, 'Machine is not using a recognized NIS server',
$CONFIGRULESFAIL, 'This machine\'s hardware does not match the rules for the specified image',
$CONFIGDIFFRP, 'Root password doesn\'t match production root password',
$CONFIGDNJ, 'do_not_jumpstart file exists on machine',
$NOBOOTSERV, 'No boot server on the same subnet as this machine',
$NOINSTALLSERV, 'No install server is configured for this image, version, etc.',
$ROGUERARP, 'There is a rogue RARP server on this machine\'s subnet',
$ROGUEBOOTPARAMS, 'There is a rogue bootparams server on this machine\'s subnet',
$CONFIGSUCCESS, 'Machine successfully configured and ready to be jumpstarted',
$STARTNOPING, 'start was unable to ping this machine',
$STARTNORSH, 'start was unable to rsh to this machine',
$USERLO, 'User logged on to machine',
$STARTDIFFRP, 'Root password doesn\'t match production root password',
$STARTDNJ, 'do_not_jumpstart file exists on machine',
$ODBSERVER, 'ODB_SERVER running on machine',
$NEEDSBNI, 'Machine ready to be jumpstarted but it is new or down at prom and can\'t be automatically started',
$STARTED, 'Machine has just been rebooted and will begin install shortly',
$JSNOPING, 'Machine rebooted and now can\'t ping machine',
$JSNORSH, 'Machine rebooted and now can\'t rsh to machine',
$SUNINSTALL, 'Machine has configured NIS and is starting install',
$PFINSTALL, 'Machine is configuring disk drives',
$PACKAGEINSTALL, 'Machine is installing packages',
$FINISHSCRIPT, 'Machine is executing finish script',
$FINALBOOT, 'Machine has finished install and is rebooting',
$JUMPFIN, 'Jumpstart finished, the machine will be ready shortly',
$TJFAIL, 'testjump failed',
$TNFAIL, 'testnet failed',
$READY, 'Machine ready to use',
);

###############
# Subroutines #
###############

###############################################################################
# Designed to be used to lock the status system.  The idea is that if a       #
# process wants to update the status of one or more hosts, it should lock the #
# system, read the status file, change the status of the hosts, write the     #
# status file back out and then unlock the status system.  The prevents two   #
# processes from reading the status file, making different changes to the     #
# status and then trying to write them back out.  Whichever one went first    #
# would be tossing its changes in the bit bucket because the second process   #
# would overwrite them.                                                       #
###############################################################################
sub superlock
{
	lockfile($SUPERLOCKFILE, $_[0]);
	unlockstat();  # Just in case...
}

#############################################################################
# Designed to lock the status file itself.  Why do we need this you ask?    #
# Isn't the superlock above sufficient?  Well, ordinarily it would be.  But #
# with our start script forking the way it does, we want it to be able to   #
# have all of the forks simultaneously making status changes to their       #
# respective hosts and then combine those changes after they finish.  As    #
# such, the start script as a whole locks the status system with superlock  #
# and then the individual forks lock the status file when they are reading  #
# and writing from it.                                                      #
#############################################################################
sub lockstat
{
	# Unless they've acknoledged the risk, make them superlock first
	unless ($_[1] == 100)
	{
		die "You must superlock before using the status file\n"
			unless (issuperlocked());
	}

	lockfile($LOCKFILE, $_[0]);
}

##################################
# Create the specified lock file #
##################################
sub lockfile
{
	my $lockfile = $_[0];
	my $output_type = $_[1];
	my $counter = 1;
	my $retrycounter = 0;

	# Define how we should output warnings/errors.  Silent says don't output
	# them, html says to output them as HTML comments.
	my $silent = 0;
	my $html = 0;
	if (defined($output_type) && $output_type == 1)
	{
		$silent = 1;
	}
	elsif (defined($output_type) && $output_type == 2)
	{
		$html = 1;
	}

	WAIT: while (-e $lockfile) 
	{ 
		$counter++;
		print "<!--" if ($counter%100 == 25 && $html);
		print "Looks like someone else is running config or start, " .
			"please wait..." if ($counter%100 == 25 && !$silent);
		print "-->" if ($counter%100 == 25 && $html);
		print "\n" if ($counter%100 == 25 && !$silent);
		sleep($LOCKWAIT); 
	}

	# We have to be careful here.  In some cases there are many processes
	# waiting to lock the same file.  As such, occasionally someone will
	# beat us to the punch here.  I.e. both processes will simultaneously
	# notice the lock file has been deleted and try to open a new one.
	# If our open fails we go back to waiting.
	$rvalue = open(LF, ">$lockfile");
	if (!$rvalue)
	{
		# Looks like someone beat us to it.  This isn't foolproof...
		# If we don't have write permissions to the lock file this
		# will just loop endlessly.  That would be be bad so if this
		# happens 100 times we die.
		die "Looks like we don't have write permissions for $lockfile"
			if ($retrycounter == 100);

		++$retrycounter;
		print "<!--" if ($html);
		print "Make sure we have write permissions to $lockfile!"
			unless ($silent);
		print "-->" if ($html);
		print "\n" unless ($silent);
		goto WAIT;
	}

	print LF "$$ on " . `/usr/bin/hostname`; 
	close(LF);

	# Make sure we really do own the lock file
	#sleep int(rand()*3);
	sleep 1;
	open(LF, $lockfile);
	my ($pid) = split(' ', <LF>);
	close(LF);
	if ($pid ne $$)
	{
		print "<!--" if ($html);
		print "My pid seems to be $$ and the lock file has $pid.  " .
			"We'll try again.\n" unless ($silent);
		print "-->" if ($html);
		print "\n" unless ($silent);
		goto WAIT;
	}
}

#################################
# Un-super-lock the status file #
#################################
sub unsuperlock
{
	unlockstat();
	unlink($SUPERLOCKFILE);
}

##########################
# Unlock the status file #
##########################
sub unlockstat
{
	unlink($LOCKFILE);
}

###############################
# Check if a lock file exists #
###############################
sub islocked
{
	if (-e $_[0])
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

########################################################
# Check whether the status file is super-locked or not #
########################################################
sub issuperlocked
{
	return islocked($SUPERLOCKFILE);
}

sub isstatlocked
{
	return islocked($LOCKFILE);
}

###########################
# Read in the status file #
###########################
sub readstat
{
	# Partialwritestat tells us to use some temporary hashes so it
	# can combine it's info with what is in the status file.
	$usetemp = $_[0];
	$outputtype = $_[1];
	$ignoresuper = $_[2];

	# Partialwritestat will already have the status file locked so
	# we can't lock it.  No problem in this case since we know it
	# is waiting for us.
	lockstat($outputtype, $ignoresuper) unless ($usetemp);

	print "    Start of readstat\n" if ($DEBUGLEVEL >= 1);

	# Make sure we don't have a zero-length status file.  Occasionally
	# the status file gets blown away by something.  If we catch that
	# and bail out then one of the backup copies we make can be restored
	# with no problems.  If we blindly push ahead then we'll overwrite
	# the backup copies and mess everything up.
	checksize();

	# Clear out hosts array
	@hosts = ();

	# One last check to make sure the status file exists
	unless (-e $STATFILE)
	{
		unlockstat();
		die "readstat: Status file doesn't exist\n";
	}
	
	# And that it is readable
	unless (-r $STATFILE)
	{
		unlockstat();
		die "readstat: Status file isn't readable\n";
	}

	open(SF, $STATFILE);
	LINE: while(<SF>)
	{
		chomp;

		my ($host, $ip, $ether, $arch, $down, $building,
                    $room, $statuslist) = split(/\t/);

		unless ($host)
		{
			print "Bad line in status file, hostname required:\n$_\n";
			next LINE;
		}

		# Set things to intelligent defaults if they are missing
		$ip = 'unknown' unless ($ip);
		$ether = 'unknown' unless($ether);
		$arch = 'unknown' unless ($arch);
		$down = 0 unless ($down);
		$building = 'unknown' unless ($building);
		$room = 'unknown' unless ($room);

		if (!defined($statuslist) || $statuslist eq '')
		{
			$statuslist = $NOTTOUCHED;
		}

		# The status codes are stored as a comma seperated list in
		# the status file but we want them in memory as an array.  So
		# we split them up here.
		my @status = split(/,/, $statuslist);
		
		unless ($usetemp)
		{
			push(@hosts, $host);
			$ip{$host} = $ip;
			$ether{$host} = $ether;
			$arch{$host} = $arch;
			$down{$host} = $down;
			$building{$host} = $building;
			$room{$host} = $room;
			$status{$host} = [ @status ];
		}
		else
		{
			# Note the different names for the hashes
			push(@temphosts, $host);
			$tempip{$host} = $ip;
			$tempether{$host} = $ether;
			$temparch{$host} = $arch;
			$tempdown{$host} = $down;
			$tempbuilding{$host} = $building;
			$temproom{$host} = $room;
			$tempstatus{$host} = [ @status ];
		}
	}
	close(SF);

	checksize();

	print "    End of readstat\n" if ($DEBUGLEVEL >= 1);

	unless ($usetemp)
	{
		unlockstat();
	}
}

####################################
# Write out an updated status file #
####################################
sub writestat
{
	lockstat();

	print "    Start of writestat\n" if ($DEBUGLEVEL >= 1);

	checksize();

	rename($STATFILE, "$STATFILE-") || die "rename failed";
	#system("cp $STATFILE /tmp");

	my $rvalue = open(SF, ">$STATFILE");
	unless ($rvalue)
	{
		unlockstat();
		die "writestat: Unable to open status file for writing\n";
	}

	my $host;
	foreach $host (@hosts)
	{
		my $statuslist = join(',', @{$status{$host}});
		print SF "$host\t$ip{$host}\t$ether{$host}\t$arch{$host}\t" .
			"$down{$host}\t$building{$host}\t$room{$host}\t$statuslist\n";
	}
	close(SF);

	print "    End of writestat\n" if ($DEBUGLEVEL >= 1);

	checksize();

	unlockstat();
}

########################################################################
# Add a given status code to the list of codes associated with a host. #
# If the code is in the same range as the current codes it is appended #
# to the list.  If it is in a different range then the list is reset   #
# to the given code.                                                   #
########################################################################
sub addstat
{
	my $host = $_[0];
	my $newstat = $_[1];

	# We can only have multiple status codes for a host if all of
	# the codes are errors/warnings and within the same stage of the
	# jumpstart process, i.e. within the same group of 100.
	if (int($status{$host}[0] / 100) == int($newstat / 100) &&
            $status{$host}[0]%100 != 0 &&
            $newstat%100 != 0)
	{
		my $alreadylisted = 0;
		STATUS: foreach $teststatus (@{$status{$host}})
		{
			if ($teststatus == $newstat)
			{
				$alreadylisted = 1;
				last STATUS;
			}
		}

		if (! $alreadylisted)
		{
			push(@{$status{$host}}, $newstat);
		}
	}
	else
	{
		setstat($host, $newstat);
	}
}

#######################################################################
# Set the status of a host to a given value, overwriting any existing #
# status values.                                                      #
#######################################################################
sub setstat
{
	my $host = $_[0];
	my $newstat = $_[1];

	$status{$host} = [ ($newstat) ];
}

########################################################################
# Return a pretty listing of all of the status codes associated with a #
# given host.  We can either format it as plain text or HTML.          #
########################################################################
sub stattotext
{
	my $host = $_[0];
	my $html = $_[1];
	my $text;
	my $first = 1;

	foreach $stat (@{$status{$host}})
	{
		if ($first)
		{
			$first = 0;
			$text = $codestring{$stat};
		}
		else
		{
			if ($html)
			{
				$text .= "<BR>";
			}
			$text .= "\n$codestring{$stat}";
		}
	}

	return $text;
}

######################################################
# Check that the size of the status file is not zero #
######################################################
sub checksize
{
	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
            $atime,$mtime,$ctime,$blksize,$blocks) = stat($STATFILE);

	print "    Statusfile is $size bytes\n" if ($DEBUGLEVEL >= 1);

	if ($size == 0)
	{
		die "Woah, statusfile blown away!!!";
	}

	# Make a backup copy
	system("cp $STATFILE /tmp");
}

sub readinputfile
{
	my $inputfile = $_[0];
	my @ihosts;

	open(INPUT, $inputfile) || die "Failed to open $inputfile as an input file";

	while(<INPUT>)
	{
		chop;
		my ($data, $comment) = split(/#/);
		my ($ihost, $ibuilding, $iroom) = split(' ', $data);

		push(@ihosts, $ihost);
		$building{$ihost} = $ibuilding;
		$room{$ihost} = $iroom;
	}
	close(INPUT);

	return(@ihosts);
}

#####################################################
# What follows is a bit of a hack, close your eyes! #
#####################################################

sub partialwritestat
{
	lockstat();
 
	print "    Start of partialwritestat\n" if ($DEBUGLEVEL >= 1);

	checksize();

	# Read in the current status to some temporary hashes.  Note that
	# we've already locked the status file.  Among other things, the
	# arg to readstat tells it to not try to lock the statusfile.
	readstat(1);
	
	rename($STATFILE, "$STATFILE-") || die "rename failed";
	my $rvalue = open(SF, ">$STATFILE");
	if (!$rvalue)
	{
		unlockstat();
		die "partialwritestat: Unable to open status file for writing\n";
	}
 
	my $host;
	foreach $host (@temphosts)
	{
		# If this is a host we are authoritative for, use our info
		if (inarray($host, @_))
		{
			my $statuslist = join(',', @{$status{$host}});
			print SF "$host\t$ip{$host}\t$ether{$host}\t$arch{$host}" .
				"\t$down{$host}\t$building{$host}\t$room{$host}\t$statuslist\n";
		}
		else  # Otherwise we use the info we read in from the file
		{
			my $tempstatuslist = join(',', @{$tempstatus{$host}});
			print SF "$host\t$tempip{$host}\t$tempether{$host}" .
				"\t$temparch{$host}\t$tempdown{$host}" .
				"\t$tempbuilding{$host}\t$temproom{$host}" .
				"\t$tempstatuslist\n";
		}
	}
	close(SF);
 
	print "    End of partialwritestat\n" if ($DEBUGLEVEL >= 1);
	checksize();

	unlockstat();
}

return 1;

